from enum import Enum
from typing import List, Optional


class Provider(Enum):
    aws = "AWS"
    azure = "AZURE"
    digital_ocean = "DIGITALOCEAN"
    equinix = "EQUINIX"
    gcp = "GCP"

    def __str__(self) -> str:
        return str(self.value)


def to_providers(providers: Optional[List[str]]) -> List[Provider]:
    if not providers:
        return []

    return [Provider(provider) for provider in providers]



from .app import App
from .build import build
from .decorators import app, context, task
from .deploy import deploy
from .entry_points import start
from .task import Task

__all__ = (
    "Task",
    "context",
    "task",
    "app",
    "start",
    "App",
    "build",
    "deploy",
)



from typing import Any, Dict, List, NamedTuple, Optional


class Flight(NamedTuple):
    """
    Flight class.
    """

    oid: str
    name: str
    image: str
    status: Optional[str] = None


def to_flights(flights: List[Dict[str, Any]]) -> List[Flight]:
    return [Flight(**flight) for flight in flights]


class Formation(NamedTuple):
    """
    Formation class.
    """

    oid: str
    name: str
    url: Optional[str] = None
    flights: List[Flight] = []
    gateway_flight: Optional[str] = None


class MetaPage(NamedTuple):
    total: int
    next: str
    prev: str


class FormationPage(NamedTuple):
    formations: List[Formation]
    meta: MetaPage


def to_formation_page(formation_page: Dict[str, Any]) -> FormationPage:
    return FormationPage(
        formations=to_formations(formation_page["objects"]),
        meta=MetaPage(**formation_page["meta"]),
    )


def to_formation(formation: Dict[str, Any]) -> Formation:
    gateway = formation.pop("gateway-flight")
    flights = to_flights(formation.pop("flights"))
    formation_modified = {**formation, "flights": flights, "gateway_flight": gateway}
    return Formation(**formation_modified)


def to_formations(formations: List[Dict[str, Any]]) -> List[Formation]:
    return [to_formation(formation) for formation in formations]



from typing import Any, Dict, NamedTuple, Optional

import simplejson as json

SDK_HTTP_ERROR_CODE = 0


def headers(api_key: Optional[str]) -> Dict[str, str]:
    return {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "Authorization": f"Bearer {api_key}",
    }


def to_json(any: NamedTuple) -> Any:
    return json.loads(json.dumps(any), object_hook=_remove_nulls)


def _remove_nulls(d: Dict[Any, Any]) -> Dict[Any, Any]:
    return {k: v for k, v in d.items() if v is not None}



from typing import Any, BinaryIO, Dict, List, NamedTuple, Optional

from ...util.base64url import base64url_decode_to_bytes


class Key(NamedTuple):
    """
    bytes Key of the key-value pair.
    """

    key: bytes


class KeyString(Key):
    """
    String Key of the key-value pair.

    Attributes
        ----------
        key : bytes
            key of the key-value pair
        key_str: str
            key of the key-value pair in string
    """

    key_str: str

    def __new__(cls, key: str, encoding: str = "utf-8", **kwargs):  # type: ignore
        return super().__new__(cls, key.encode(encoding), **kwargs)

    def __init__(self, key: str):
        self.key_str = key


class KeyValue(NamedTuple):
    """
    bytes Key Value class.

    Attributes
        ----------
        key : bytes
            key of the key-value pair.
        value: bytes
            value of the key-value pair.
    """

    key: bytes
    value: bytes


class KeyValueString(KeyValue):
    """
    String Key Value class.

    Attributes
        ----------
        key : bytes
            key of the key-value pair.
        value: bytes
            value of the key-value pair.
        key_str : str
            key of the key-value pair in String
        value_str: str
            value of the key-value pair in String
    """

    key_str: str
    value_str: str

    def __new__(cls, key, value, encoding="utf-8", **kwargs):  # type: ignore
        return super().__new__(cls, key.encode(encoding), value.encode(encoding), **kwargs)

    def __init__(self, key: str, value: str):
        self.key_str = key
        self.value_str = value


class KeyValueStream(KeyValue):
    """
    bytes Key and IO stream Value class.

    Attributes
        ----------
        key : bytes
            key of the key-value pair.
        value: BinaryIO
            value of the key-value.
    """

    value_stream: BinaryIO

    def __new__(cls, key, value, **kwargs):  # type: ignore
        content: bytes = b""

        with value as binary_stream:
            content = binary_stream.read()

        return super().__new__(cls, key, content, **kwargs)

    def __init__(self, key: bytes, value: BinaryIO):
        self.value_stream = value


class KeyValuePage(NamedTuple):
    """
    bytes Key Value Page class.

    It contains a paginated list of key-value, next_key is used for the next call.

    Attributes
        ----------
        key_value_pairs : KeyValue
            list of key-value
        next_key: Optional[Key]
            If next_key is non-null and you want to get the next KeyValuePage,
            you can repeat a query using next_key to continue getting KeyValuePage.
    """

    key_value_pairs: List[KeyValue]
    next_key: Optional[Key]


def to_key_value_page(keyvalue_range: Dict[str, Any]) -> KeyValuePage:
    return KeyValuePage(
        key_value_pairs=[to_key_value(key_value) for key_value in keyvalue_range["kvs"]],
        next_key=_to_key(keyvalue_range["next_key"]),
    )


def _to_key(key: Optional[str]) -> Optional[Key]:
    if key is None:
        return None

    return Key(base64url_decode_to_bytes(key))


def to_key_value(key_value: Dict[str, Any]) -> KeyValue:
    return KeyValue(
        key=base64url_decode_to_bytes(key_value["key"]),
        value=base64url_decode_to_bytes(key_value["value"]),
    )


def to_key_value_string(key_value: KeyValue, encoding: str = "utf-8") -> KeyValueString:
    return KeyValueString(
        key=key_value.key.decode(encoding), value=key_value.value.decode(encoding)
    )



from importlib.metadata import version

from .api.compute_api import ComputeAPI
from .api.lock_api import LockAPI
from .api.metadata_api import MetadataAPI
from .api.restrict_api import RestrictAPI
from .api.sql_api import GlobalSQL
from .api.token_api import TokenAPI
from .apps import App, Task, app, build, context, start, task
from .configuration import Configuration, config
from .logging import log

__version__ = version(__name__)


class Seaplane:
    @property
    def config(self) -> Configuration:
        return config

    @property
    def auth(self) -> TokenAPI:
        return config._token_api

    @property
    def metadata(self) -> MetadataAPI:
        return MetadataAPI(config)

    @property
    def locks(self) -> LockAPI:
        return LockAPI(config)

    @property
    def restrict(self) -> RestrictAPI:
        return RestrictAPI(config)

    @property
    def compute(self) -> ComputeAPI:
        return ComputeAPI(config)

    @property
    def global_sql(self) -> GlobalSQL:
        return GlobalSQL(config)


sea = Seaplane()

__all__ = [
    "App",
    "Task",
    "task",
    "app",
    "start",
    "log",
    "context",
    "build",
]



import asyncio
import functools
import json
import os
import sys
import traceback
from typing import Any, List, Optional
from urllib.parse import urlparse
import uuid

from flask import Flask, jsonify, request
from flask_cors import CORS
from flask_socketio import SocketIO, emit
import nats

from ...carrier import processor
from ...configuration import config
from ...logging import log
from ...model.errors import SeaplaneError
from ..app import App
from ..build import build
from ..datasources import RequestDataSource
from ..decorators import apps_json, context
from ..deploy import deploy, destroy

SEAPLANE_APPS_CORS = list(os.getenv("SEAPLANE_APPS_CORS", "http://localhost:3000").split(" "))
AUTH_TOKEN = os.getenv("SEAPLANE_APPS_AUTH_TOKEN")

app = Flask(__name__)

CORS(app, origins=SEAPLANE_APPS_CORS)

if not config.is_production():

    sio = SocketIO(app, cors_allowed_origins=SEAPLANE_APPS_CORS, async_mode="threading")

    @sio.on("message")  # type: ignore
    def handle_message(data: Any) -> None:
        ...

    @sio.on("connect")  # type: ignore
    def on_connect() -> None:
        emit("message", apps_json(context.apps))


def send_something(data: Any) -> None:
    emit("message", data, sid="sp", namespace="", broadcast=True)


def authenticate_token() -> bool:
    token = request.headers.get("Authorization")
    if token and token.startswith("Bearer "):

        if token == f"Bearer {AUTH_TOKEN}":
            return True

    return False


@app.before_request
def before_request() -> Any:
    if request.path == "/healthz":
        return

    if os.getenv("SEAPLANE_APPS_AUTH_TOKEN") is not None:
        if not authenticate_token():
            return {"message": "Unauthorized"}, 401


def dev_https_api_start() -> Flask:
    log.debug(f"CORS enabled: {SEAPLANE_APPS_CORS}")
    context.set_event(lambda data: send_something(data))

    apps = context.apps

    for myapp in apps:

        def endpoint_func(pipe: App = myapp) -> Any:
            if request.method == "POST" or request.method == "PUT":
                data = request.get_json()
                result = pipe.func(data)
                return result
            elif request.method == "GET":
                return pipe.func("nothing")  # current limitation, it needs to pass something

        endpoint = functools.partial(endpoint_func, pipe=myapp)
        app.add_url_rule(myapp.path, myapp.id, endpoint, methods=[myapp.method])

    def health() -> str:
        emit("message", {"data": "test"}, sid="lol", namespace="", broadcast=True)
        return "Seaplane Apps Demo"

    app.add_url_rule("/", "healthz", health, methods=["GET"])

    if not config.is_production():
        log.info("🚀 Apps DEVELOPMENT MODE")
        sio.run(app, debug=False, port=1337)
    else:
        log.info("🚀 Apps PRODUCTION MODE")

    return app


loop = asyncio.get_event_loop()


async def publish_message(
    stream: str, id: str, message: Any, order: int, tasks: List[str]
) -> None:
    nc = await nats.connect(
        ["nats://148.163.201.1:2003"],
        user_credentials=os.path.abspath("./carrier.creds"),
    )
    js = nc.jetstream()

    nats_message = {"id": id, "input": message, "order": order}

    for task in tasks:
        log.debug(f"Sending to {stream}.{task} task,  message: {nats_message}")
        ack = await js.publish(f"{stream}.{task}", str(json.dumps(nats_message)).encode())
        log.debug(f"ACK: {ack}")

    await nc.close()


def generate_id() -> str:
    random_id = str(uuid.uuid4())
    return random_id


def build_requests_datasource() -> RequestDataSource:
    database = os.getenv("SEAPLANE_TENANT_DB__DATABASE", "")
    username = os.getenv("SEAPLANE_TENANT_DB_USERNAME", "")
    password = os.getenv("SEAPLANE_TENANT_DB_PASSWORD", "")
    port = os.getenv("SEAPLANE_TENANT_DB_PORT", 5432)
    host = os.getenv("SEAPLANE_TENANT_DB_HOST", urlparse(config.global_sql_endpoint).netloc)
    return RequestDataSource(
        database=database, host=host, username=username, password=password, port=int(port)
    )


def prod_https_api_start() -> Flask:
    log.debug(f"CORS enabled: {SEAPLANE_APPS_CORS}")

    project = build()
    schema = project["schema"]

    context.set_event(lambda data: send_something(data))
    requests_datasource = build_requests_datasource()

    apps = context.apps

    for myapp in apps:

        def endpoint_func(pipe: App = myapp) -> Any:
            if request.method == "POST" or request.method == "PUT":
                body = request.get_json()

                if not ("input" in body):
                    return jsonify({"error": "Invalid JSON"}), 401

                id = generate_id()
                app_first_tasks = schema["apps"][pipe.id]["io"]["entry_point"]
                metadata = body["params"]

                batch = body["input"]

                for idx, content in enumerate(batch):
                    content["_params"] = metadata
                    loop.run_until_complete(
                        publish_message(pipe.id, id, content, idx, app_first_tasks)
                    )

                requests_datasource.save_request(id, len(batch))

                return jsonify({"id": id, "status": "processing"}), 200
            elif request.method == "GET":
                return pipe.func("nothing")  # current limitation, it needs to pass something

        def get_result(id: str) -> Any:

            output = None
            req_status = requests_datasource.get_request_status(id)
            if not req_status:
                return jsonify({"error": f"Request ID: {id} not found"}), 404

            batch_count = req_status["batch_count"]
            result_count = req_status["result_count"]

            if batch_count > result_count:
                status = f"{result_count}/{batch_count}"
            else:
                status = "completed"
                output = requests_datasource.get_request_results(id)

            return jsonify({"id": id, "status": status, "output": output}), 200

        endpoint = functools.partial(endpoint_func, pipe=myapp)
        app.add_url_rule(myapp.path, myapp.id, endpoint, methods=[myapp.method])
        app.add_url_rule(f"{myapp.path}/<id>", f"{myapp.id}_query", get_result, methods=["GET"])

    def health() -> str:
        return "Seaplane Apps Demo"

    app.add_url_rule("/", "healthz", health, methods=["GET"])

    if not config.is_production():
        log.info("🚀 Seaplane Apps DEVELOPMENT MODE")
        sio.run(app, debug=False, port=1337)
    else:
        log.info("🚀 Seaplane Apps PRODUCTION MODE")

    return app


def start_task(task_id: str, save_result: bool) -> None:
    task = context.get_task(task_id)

    if not task:
        raise SeaplaneError(
            f"Task {task_id} not found, \
                            make sure the Task ID is correct."
        )

    processor.start()

    requests_datasource = None
    if save_result:
        requests_datasource = build_requests_datasource()

    while True:
        try:
            log.info(f" Task {task.id}  waiting for getting data...")
            message = json.loads(processor.read())
            log.debug(f" Message recieved: {message}")

            id = message["id"]
            order = message["order"]

            try:
                message["output"] = task.process(message["input"])

                log.debug(f" Task Result: {message}")

                next_message = {"input": message["output"], "id": id, "order": order}
                processor.write(str(json.dumps(next_message)).encode())
                processor.flush()

                if requests_datasource:
                    requests_datasource.save_result(id, order, message["output"])

            except Exception as e:
                error_str = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
                log.error(
                    f"Error running Task:\
                        \n {error_str}"
                )
                next_message = {"input": {"error": error_str}, "id": id, "order": order}
                processor.write(str(json.dumps(next_message)).encode())
                processor.flush()
        except Exception as ex:
            error_str = "\n".join(traceback.format_exception(type(ex), ex, ex.__traceback__))
            log.error(
                f"Error running Task:\
                    \n {error_str}"
            )
            next_message = {"input": {"error": error_str}, "id": id, "order": order}
            processor.write(str(json.dumps(next_message)).encode())
            processor.flush()


def start() -> Optional[Flask]:
    if len(sys.argv) > 1:
        command = sys.argv[1]

        if command == "build":
            build()
            return None
        elif command == "deploy":
            if len(sys.argv) == 3:
                co_id = sys.argv[2]
                deploy(co_id)
            else:
                deploy()
            return None
        elif command == "destroy":
            destroy()
            return None

    task_id: Optional[str] = os.getenv("TASK_ID")

    if not task_id:
        log.info("Starting API Entry Point...")
        if not config.is_production():
            return dev_https_api_start()
        else:
            return prod_https_api_start()
    else:
        log.info(f"Starting Task {task_id} ...")
        save_result = os.getenv("SAVE_RESULT_TASK", "").lower() == "true"
        start_task(task_id, save_result)
        return None



from typing import Any, Callable

import requests
from requests import Response
from returns.result import Failure, Result, Success

from ..logging import log
from ..model.errors import HTTPError
from .api_http import SDK_HTTP_ERROR_CODE
from .token_api import TokenAPI


def handle_request(request: Callable[[], Response]) -> Result[Any, HTTPError]:
    try:
        response = request()

        if response.ok:
            return Success(response.json())
        else:
            body_error = response.text
            log.error(f"Request Error: {body_error}")
            return Failure(HTTPError(response.status_code, body_error))
    except requests.exceptions.RequestException as err:
        log.error(f"Request exception: {str(err)}")
        return Failure(HTTPError(SDK_HTTP_ERROR_CODE, str(err)))


def provision_req(
    token_api: TokenAPI,
) -> Callable[[Callable[[str], Response]], Result[Any, HTTPError]]:
    """
    Before every request, we make sure we use a valid access token.
    """

    def handle_request(request: Callable[[str], Response], token: str) -> Result[Any, HTTPError]:
        try:
            response = request(token)

            if response.ok:
                data = None
                if response.headers.get("content-type") == "application/json":
                    data = response.json()
                else:
                    data = response.text

                return Success(data)
            else:
                body_error = response.text
                log.error(f"Request Error: {body_error}")
                return Failure(HTTPError(response.status_code, body_error))
        except requests.exceptions.RequestException as err:
            log.error(f"Request exception: {str(err)}")
            return Failure(HTTPError(SDK_HTTP_ERROR_CODE, str(err)))

    def renew_if_fails(
        token_api: TokenAPI, request: Callable[[str], Response], http_error: HTTPError
    ) -> Result[Any, HTTPError]:
        if http_error.status != 401:
            return Failure(http_error)

        if token_api.auto_renew:
            log.info("Auto-Renew, renewing the token...")
            token = token_api.renew_token()
            return handle_request(request, token)
        else:
            return Failure(http_error)

    def req(request: Callable[[str], Response]) -> Result[Any, HTTPError]:
        access_token: Result[str, HTTPError]

        if token_api.access_token is not None:
            access_token = Success(token_api.access_token)
        else:
            access_token = token_api._request_access_token().map(lambda result: result["token"])

        return access_token.bind(lambda token: handle_request(request, token)).lash(
            lambda error: renew_if_fails(token_api, request, error)
        )

    return req



import functools
import json
from typing import Any, Callable, Dict, List, Optional

from ..logging import log
from ..model.errors import HTTPError
from .app import App, AppEvent
from .event_handler import EventHandler
from .executor import RealTaskExecutor, TaskExecutor
from .task import Task, TaskEvent


def task_to_json(task: Task) -> Dict[str, Any]:
    return {"id": task.id, "type": task.type, "model": task.model}


def app_to_json(app: App) -> Dict[str, Any]:
    return {
        "id": app.id,
        "path": app.path,
        "method": app.method,
        "tasks": [task_to_json(task) for task in app.tasks],
    }


def apps_json(apps: List[App]) -> Dict[str, Any]:
    return {
        "type": "apps",
        "payload": [app_to_json(app) for app in apps],
    }


class Context:
    def __init__(
        self,
        apps: Optional[List[App]] = None,
        tasks: Optional[List[Task]] = None,
    ) -> None:
        self.actual_app_index = -1
        self.event_handler = EventHandler()
        self.task_executor: TaskExecutor = RealTaskExecutor(self.event_handler)

        if apps is None:
            self.apps = []
        else:
            self.apps = apps

        if tasks is None:
            self.tasks = []
        else:
            self.tasks = tasks

    def active_app(self, id: str) -> None:
        for i, app in enumerate(self.apps):
            if app.id == id:
                self.actual_app_index = i
                break

    def set_event(self, callback: Callable[[Dict[str, Any]], None]) -> None:
        self.event_handler.set_event(callback)

    def add_event(self, event: AppEvent) -> None:
        self.event_handler.add_event(event)

    def update_event(self, event: AppEvent) -> None:
        self.event_handler.update_event(event)

    def task_event(self, task_event: TaskEvent) -> None:
        self.event_handler.task_event(task_event)

    def get_actual_app(self) -> Optional[App]:
        if self.actual_app_index == -1:
            return None

        return self.apps[self.actual_app_index]

    def add_app(self, app: App) -> None:
        if len(self.apps) == 1 and self.apps[0].id == "temporal":
            self.apps[0] = app
        else:
            self.actual_app_index = len(self.apps)
            self.apps.append(app)

        log.debug(f"🧠 App: {app.id}, Path: {app.path}")
        self.event_handler.on_change(apps_json(self.apps))

    def add_task(self, task: Task) -> None:
        log.debug(f"⌛️ Task {task.id} of type: {task.type}")
        self.tasks.append(task)

    def get_task(self, id: str) -> Optional[Task]:
        for c in self.tasks:
            if c.id == id:
                return c

        return None

    def assign_to_active_app(self, task: Task) -> None:
        self.apps[self.actual_app_index].add_task(task)
        app = context.get_actual_app()
        if app is not None:
            log.info(f"⌛️ Assign Task {task.id} to App: {app.id}")
        else:
            log.info(
                f"🔥 Actual App is None, can't assign \
                    Task {task.id} to App"
            )
        self.event_handler.on_change(apps_json(self.apps))

    def set_executor(self, executor: TaskExecutor) -> None:
        self.task_executor = executor


context = Context()


def app(
    path: str,
    id: str,
    method: str = "POST",
    parameters: Optional[List[str]] = None,
    _func: Optional[Callable[[Any], Any]] = None,
) -> Callable[[Any, Any], Any]:
    def decorator_apps(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            log.debug(f"App Path: {path}, Method: {method}, ID: {id}")
            context.active_app(id)

            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = AppEvent(app_id=id, input=args_json)
            context.add_event(event)
            result = None

            try:
                result = func(*args, **kwargs)
                event.set_output(result)
            except HTTPError as err:
                log.error(f"App error: {err}")
                event.set_error(err)

            context.update_event(event)
            return result

        app = App(wrapper, path, method, id, parameters)
        context.add_app(app)
        return wrapper

    if not _func:
        return decorator_apps  # type: ignore
    else:
        return decorator_apps(_func)


def import_task(_func: Optional[Callable[[Any], Any]], task: Task) -> Callable[[Any, Any], Any]:
    def decorator_task(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = TaskEvent(task.id, args_json)
            context.task_event(event)

            result = task.process(*args, **kwargs)

            event.set_output(result)
            context.task_event(event)

            return func(result)

        return wrapper

    context.add_task(task)

    if not _func:
        return decorator_task  # type: ignore
    else:
        return decorator_task(_func)


def task(
    type: str,
    id: Optional[str] = None,
    model: Optional[str] = None,
    sql: Optional[Dict[str, str]] = None,
    _func: Optional[Callable[[Any], Any]] = None,
) -> Callable[[Any, Any], Any]:
    def decorator_task(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:

        task = Task(func=func, type=type, model=model, id=id, sql_access=sql)
        context.add_task(task)

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            context.assign_to_active_app(task)

            return context.task_executor.execute(task, *args, **kwargs)

        return wrapper

    if not _func:
        return decorator_task  # type: ignore
    else:
        return decorator_task(_func)



from typing import Any, Dict, List, NamedTuple, Optional

from ..provider import Provider, to_providers
from ..region import Region, to_regions
from .flight import Flight, to_flights


class FormationConfiguration(NamedTuple):
    """
    Formation class with all formation attributes.
    """

    flights: List[Flight]
    affinity: Optional[List[str]] = None
    connections: Optional[List[str]] = None
    public_endpoints: Optional[object] = None
    formation_endpoints: Optional[object] = None
    flight_endpoints: Optional[object] = None
    providers_allowed: Optional[List[Provider]] = None
    providers_denied: Optional[List[Provider]] = None
    regions_allowed: Optional[List[Region]] = None
    regions_denied: Optional[List[Region]] = None


def to_formation_config(fconfig: Dict[str, Any]) -> FormationConfiguration:
    formation_config = fconfig.copy()
    formation_config.pop("flights", None)
    formation_config.pop("providers_allowed", None)
    formation_config.pop("providers_denied", None)
    formation_config.pop("regions_allowed", None)
    formation_config.pop("regions_denied", None)

    return FormationConfiguration(
        **formation_config,
        flights=to_flights(fconfig["flights"]),
        providers_allowed=to_providers(fconfig.pop("providers_allowed", None)),
        providers_denied=to_providers(fconfig.pop("providers_denied", None)),
        regions_allowed=to_regions(fconfig.pop("regions_allowed", None)),
        regions_denied=to_regions(fconfig.pop("regions_denied", None)),
    )



from typing import Any, List, Text

import requests
from returns.result import Result

from ..configuration import Configuration, config
from ..model.compute.active_configuration import ActiveConfiguration
from ..model.compute.formation_configuration import FormationConfiguration, to_formation_config
from ..model.errors import HTTPError
from .api_http import headers, to_json
from .api_request import provision_req


class FormationConfigurationAPI:
    """
    Class for handle Configuration and Active Configuration API calls.
    Links:
      - https://developers.seaplane.io/reference/get_formations-formationname-configurations
      - https://developers.seaplane.io/reference/get_formations-formationname-activeconfiguration
    """

    def __init__(self, configuration: Configuration = config) -> None:
        self.url = f"{configuration.compute_endpoint}/formations"
        self.req = provision_req(configuration._token_api)

    def create(
        self, formation_name: str, formation: FormationConfiguration, active: bool = False
    ) -> Result[str, HTTPError]:
        url = f"{self.url}/{formation_name}/configurations"
        payload = to_json(formation)
        params = {"active": active}

        return self.req(
            lambda access_token: requests.post(
                url=url, json=payload, params=params, headers=headers(access_token)
            )
        )

    def get_all(self, formation_name: Text) -> Result[List[str], HTTPError]:
        url = f"{self.url}/{formation_name}/configurations"
        return self.req(lambda access_token: requests.get(url, headers=headers(access_token)))

    def get(self, formation_name: Text, id: Text) -> Result[FormationConfiguration, HTTPError]:
        url = f"{self.url}/{formation_name}/configurations/{id}"

        return self.req(lambda access_token: requests.get(url, headers=headers(access_token))).map(
            lambda json: to_formation_config(json)
        )

    def delete(self, formation_name: Text, id: Text) -> Result[Any, HTTPError]:
        url = f"{self.url}/{formation_name}/configurations/{id}"

        return self.req(lambda access_token: requests.delete(url, headers=headers(access_token)))

    def get_active_config(self, formation_name: Text) -> Result[Any, HTTPError]:
        url = f"{self.url}/{formation_name}/activeConfiguration"

        return self.req(lambda access_token: requests.get(url, headers=headers(access_token)))

    def set_active_config(
        self, formation_name: Text, active_configuration: ActiveConfiguration, force: bool
    ) -> Result[Any, HTTPError]:
        url = f"{self.url}/{formation_name}/activeConfiguration"
        params = {"force": force}
        payload = active_configuration.__dict__

        return self.req(
            lambda access_token: requests.put(
                url, headers=headers(access_token), params=params, json=payload
            )
        )

    def stop_formation(self, formation_name: Text) -> Result[Any, HTTPError]:
        url = f"{self.url}/{formation_name}/activeConfiguration"

        return self.req(lambda access_token: requests.delete(url, headers=headers(access_token)))



from enum import Enum
from typing import List, Optional


class Region(Enum):
    asia = "XA"
    republic_of_china = "XC"
    europe = "XE"
    africa = "XF"
    north_america = "XN"
    oceania = "XO"
    antartica = "XQ"
    south_america = "XS"
    uk = "XU"

    def __str__(self) -> str:
        return str(self.value)


def to_regions(regions: Optional[List[str]]) -> List[Region]:
    if not regions:
        return []

    return [Region(region) for region in regions]



from typing import Any, Dict, List, Optional

import requests

from ..configuration import Configuration, config
from ..model.locks import HeldLock, Lock, LockPage, Name, to_held_lock, to_lock, to_lock_page
from ..util import unwrap
from ..util.base64url import base64url_encode_from_bytes
from .api_http import headers
from .api_request import provision_req


class LockAPI:
    """
    Class for handle Lock API calls.
    """

    def __init__(self, configuration: Configuration = config) -> None:
        self.url = f"{configuration.coordination_endpoint}/locks"
        self.req = provision_req(configuration._token_api)

    def get(self, name: Name) -> Lock:
        """Gets information about a single lock.

        Parameters
        ----------
        name : Name
            Name of the lock.

        Returns
        -------
        Lock
            Returns Lock if successful or it will raise an HTTPError otherwise.
        """
        _url = f"{self.url}/base64:{base64url_encode_from_bytes(name.name)}"

        return unwrap(
            self.req(lambda access_token: requests.get(_url, headers=headers(access_token))).map(
                lambda lock_response: to_lock(lock_response)
            )
        )

    def acquire(self, name: Name, client_id: str, ttl: int) -> HeldLock:
        """Attempts to acquire the lock with the given lock name with the given TTL.
        Client-ID should identify the client making the request for debugging purposes.

        Parameters
        ----------
        name : Name
            Name of the lock.
        client_id : Name
            client_id is an identifier showing who is currently holding the lock.
        ttl: int
            This is the requested time to live, in seconds.

        Returns
        -------
        HeldLock
            Returns HeldLock if successful or it will raise an HTTPError otherwise.
        """
        _url = f"{self.url}/base64:{base64url_encode_from_bytes(name.name)}"

        params: Dict[str, Any] = {}
        params["client-id"] = client_id
        params["ttl"] = ttl

        return unwrap(
            self.req(
                lambda access_token: requests.post(
                    _url, params=params, headers=headers(access_token)
                )
            ).map(lambda held_lock_response: to_held_lock(held_lock_response))
        )

    def release(self, name: Name, id: str) -> bool:
        """Attempts to release the given lock.

        Parameters
        ----------
        name : Name
            Name of the lock.
        id: str
            ID which allows for the renewal/release of a lock, represents "holding"
            the lock, this is usually only seen by the original acquirer of the lock
            but may be leaked by listing the locks.

        Returns
        -------
        bool
            Returns bool if successful or it will raise an HTTPError otherwise.
        """
        _url = f"{self.url}/base64:{base64url_encode_from_bytes(name.name)}"

        params: Dict[str, Any] = {}
        params["id"] = id

        release_result = unwrap(
            self.req(
                lambda access_token: requests.delete(
                    _url, params=params, headers=headers(access_token)
                )
            )
        )

        return bool(release_result == "OK")  # mypy bug, mypy can't know the

    def renew(self, name: Name, id: str, ttl: int) -> bool:
        """Attempts to renew the given lock.

        Parameters
        ----------
        name : Name
            Name of the lock.
        id: str
            ID which allows for the renewal/release of a lock, represents "holding"
            the lock, this is usually only seen by the original acquirer of the lock
            but may be leaked by listing the locks.
        ttl: int
            This is the requested time to live, in seconds.

        Returns
        -------
        bool
            Returns bool if successful or it will raise an HTTPError otherwise.
        """
        _url = f"{self.url}/base64:{base64url_encode_from_bytes(name.name)}"

        params: Dict[str, Any] = {}
        params["id"] = id
        params["ttl"] = ttl

        renew_result = unwrap(
            self.req(
                lambda access_token: requests.patch(
                    _url, params=params, headers=headers(access_token)
                )
            )
        )

        return bool(renew_result == "OK")

    def get_page(
        self,
        directory: Optional[Name] = None,
        from_lock: Optional[Name] = None,
    ) -> LockPage:
        """Returns a single page of lock information for the given directory,
        beginning with the `from_lock` key.

        If no directory is given, the root directory is used.
        If no `from_lock` is given, the range begins from the start.

        Parameters
        ----------
        directory : Optional[Name]
            The metadata key-value store supports the use of directories.
            Directories are important tools to set up region and provider restrictions
            for groups of key-value pairs.
        from_lock : Optional[Name]
            If more pages are desired, perform another range request using
            the `from_lock` value from the first request as the `from_lock` value of
            the following request.

        Returns
        -------
        LockPage
            Returns LockPage if successful or it will raise an HTTPError otherwise.
        """

        _url = self.url

        if directory is not None:
            _url = f"{self.url}/base64:{base64url_encode_from_bytes(directory.name)}/"

        params: Dict[str, Any] = {}
        if from_lock is not None:
            params["from"] = f"base64:{base64url_encode_from_bytes(from_lock.name)}"

        return unwrap(
            self.req(
                lambda access_token: requests.get(
                    _url, params=params, headers=headers(access_token)
                )
            ).map(lambda lock_range: to_lock_page(lock_range))
        )

    def get_all_pages(
        self,
        directory: Optional[Name] = None,
        from_lock: Optional[Name] = None,
    ) -> List[Lock]:
        """Returns all held lock information for the given directory,
        from the `from_lock` key onwards. May perform multiple requests.

        If no directory is given, the root directory is used.
        If no `from_lock` is given, the range begins from the start.

        Parameters
        ----------
        directory : Optional[Name]
            The metadata key-value store supports the use of directories.
            Directories are important tools to set up region and provider restrictions
            for groups of key-value pairs.
        from_lock : Optional[Name]
            It begins from from_lock in advance until the last page,
            if you don't want to begin from the start.

        Returns
        -------
        List[Lock]
            Returns a List of Locks if successful or it will raise an HTTPError otherwise.
        """

        pages: List[Lock] = []
        _from_lock = from_lock

        while True:
            page_result = self.get_page(directory, _from_lock)

            page: LockPage = page_result
            pages.extend(page.locks)

            if page.next_lock is not None:
                _from_lock = page.next_lock
            else:
                return pages



from typing import Any, Dict, List, Optional

import requests

from ..configuration import Configuration, config
from ..model.errors import SeaplaneError
from ..model.metadata import Key
from ..model.restrict import (
    Restriction,
    RestrictionDetails,
    RestrictionPage,
    SeaplaneApi,
    to_restriction,
    to_restriction_page,
)
from ..util import unwrap
from ..util.base64url import base64url_encode_from_bytes
from .api_http import headers
from .api_request import provision_req


class RestrictAPI:
    """
    Class for handle Restrict API calls.
    """

    def __init__(self, configuration: Configuration = config) -> None:
        self.url = f"{configuration.coordination_endpoint}/restrict"
        self.req = provision_req(configuration._token_api)

    def get(self, api: SeaplaneApi, key: Key) -> Restriction:
        """Returns restriction details for an API-directory combination

        Parameters
        ----------
        api: SeaplaneApi
            The name of a Seaplane API,
              for example "SeaplaneApi.metadata" or "SeaplaneApi.locks"
        key: Key
            Key pointing to a directory.

        Returns
        -------
        Restriction
            Returns Restriction if successful or it will raise an HTTPError otherwise.
        """
        _url = f"{self.url}/{api}/base64:{base64url_encode_from_bytes(key.key)}/"

        return unwrap(
            self.req(lambda access_token: requests.get(_url, headers=headers(access_token))).map(
                lambda restriction_response: to_restriction(restriction_response)
            )
        )

    def set(self, api: SeaplaneApi, key: Key, restriction_details: RestrictionDetails) -> bool:
        """Sets a restriction for an API-directory combination

        Parameters
        ----------
        api: SeaplaneApi
            The name of a Seaplane API,
              for example "SeaplaneApi.metadata" or "SeaplaneApi.locks"
        key: Key
            Key pointing to a directory.
        restriction_details: RestrictionDetails
            Allow or deny lists for geographies and cloud providers
            to be associated with a given directory.

        Returns
        -------
        Restriction
            Returns Restriction if successful or it will raise an HTTPError otherwise.
        """
        _url = f"{self.url}/{api}/base64:{base64url_encode_from_bytes(key.key)}/"

        _payload = {
            "regions_allowed": [str(region) for region in restriction_details.regions_allowed],
            "regions_denied": [str(region) for region in restriction_details.regions_denied],
            "providers_allowed": [
                str(provider) for provider in restriction_details.providers_allowed
            ],
            "providers_denied": [
                str(provider) for provider in restriction_details.providers_denied
            ],
        }

        set_result = unwrap(
            self.req(
                lambda access_token: requests.put(
                    _url, json=_payload, headers=headers(access_token)
                )
            )
        )

        return bool(set_result == "Ok")  # mypy bug, mypy can't know the type

    def delete(
        self,
        api: SeaplaneApi,
        key: Key,
    ) -> bool:
        """Removes a restriction for an API-directory combination

        Parameters
        ----------
        api: SeaplaneApi
            The name of a Seaplane API,
              for example "SeaplaneApi.metadata" or "SeaplaneApi.locks"
        key: Key
            Key pointing to a directory.

        Returns
        -------
        boolean
            Returns true if successful or it will raise an HTTPError otherwise.
        """
        url = f"{self.url}/{api}/base64:{base64url_encode_from_bytes(key.key)}/"

        delete_result = unwrap(
            self.req(lambda access_token: requests.delete(url, headers=headers(access_token)))
        )

        return bool(delete_result == "Ok")  # mypy bug, mypy can't know the

    def get_page(
        self,
        api: Optional[SeaplaneApi] = None,
        from_restriction: Optional[Key] = None,
        is_all_range: bool = False,
    ) -> RestrictionPage:
        """Returns a single page of restriction information for the given Seaplane API,
        beginning with the `from_restriction` key.

        If more pages are desired, perform another page request using the
        `next_api` and `next_key` values from the first `RestrictionPage` request as the
        `api` and `from_restriction` values of the following request, or use `get_all_pages`.


        Parameters
        ----------
        api: Optional[SeaplaneApi]
            The name of a Seaplane API,
              for example "SeaplaneApi.metadata" or "SeaplaneApi.locks"
        from_restriction : Optional[Key]
            If more pages are desired, perform another range request using
            the `from_restriction` value from the first request as the `from_restriction` value of
            the following request. If you use `from_restriction` is mandatory to indicate
            the `api` as well.
        is_all_range: bool
            Set to True to get all Restrictions from any Seaplane API, It filters by API otherwise.
            By default `is_all_range` is False. If you pass `api` parameter
            will be used to filter by API.

        Returns
        -------
        RestrictionPage
            Returns RestrictionPage if successful or it will raise an HTTPError otherwise.
        """

        _url = self.url

        if api is not None and not is_all_range:
            _url = f"{self.url}/{api}"

        params: Dict[str, Any] = {}
        if from_restriction is not None:
            params["from"] = f"base64:{base64url_encode_from_bytes(from_restriction.key)}"
            params["from_api"] = str(api)
            if not api:
                raise SeaplaneError("You must set api with from_restriction parameters.")

        return unwrap(
            self.req(
                lambda access_token: requests.get(
                    _url, params=params, headers=headers(access_token)
                )
            ).map(lambda lock_range: to_restriction_page(lock_range))
        )

    def get_all_pages(
        self,
        api: Optional[SeaplaneApi] = None,
        from_restriction: Optional[Key] = None,
    ) -> List[Restriction]:
        """
        Returns restrictions for the given Seaplane API,
        from the `from_restriction` key onwards. May perform multiple requests.

        If no API is given, it will return all API restrictions.
        If no `from_restriction` is given, the range begins from the start.

        Parameters
        ----------
        api: Optional[SeaplaneApi]
            The name of a Seaplane API,
              for example "SeaplaneApi.metadata" or "SeaplaneApi.locks"
        from_restriction : Optional[Key]
            It begins from a restriction in advance until the last page,
            if you don't want to begin from the start.

        Returns
        -------
        List[Restriction]
            Returns a List of Restrictions if successful or it will raise an HTTPError otherwise.
        """

        pages: List[Restriction] = []
        _from_restriction = from_restriction
        _from_api = api

        while True:
            page_result = self.get_page(_from_api, _from_restriction, is_all_range=True)

            page: RestrictionPage = page_result
            pages.extend(page.restrictions)

            if page.next_key is not None and page.next_api is not None:
                _from_restriction = page.next_key
                _from_api = page.next_api
            else:
                return pages



class SeaplaneError(Exception):
    """Seaplane Generic error"""

    pass


class HTTPError(SeaplaneError):
    """HTTP Errors exception

    Attributes
        ----------
        status : int
            HTTP Status code.
        message: str
            Error message.
    """

    def __init__(self, status: int, message: str = ""):
        self.status = status
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"status={self.status}, message='{self.message}'"
